----------------------------------------------------------------
-- Copyright (c) 2015 Klei Entertainment Inc.
-- All Rights Reserved.
-- SPY SOCIETY.
----------------------------------------------------------------

local util = include( "modules/util" )
local filesystem = include("modules/filesystem")

local MOD_FOLDER = "mods"

----------------------------------------------------------------
-- Mod handling

local mod_manager = class()

function mod_manager:init()
    self.mods = {}
    self:enumerateMods()

    filesystem.mountVirtualDirectory( "data_locale", "data" ) -- By default, simply map directly to data.

    local settings = savefiles.getSettings( "settings" )
    if settings.data.localeMod then
        self:mountLanguageMod( settings.data.localeMod )
    end
end

function mod_manager:updateMods()
    if not KLEISteamWorkshop then
        return
    end

    log:write( "Updating mods..." )
    KLEISteamWorkshop:setListener( KLEISteamWorkshop.EVENT_REFRESH_COMPLETE,
        function( success, msg )
            log:write("KLEISteamWorkshop.EVENT_REFRESH_COMPLETE - (%s, %s)", success and "succeeded" or "failed", msg )
            self:enumerateMods()
        end )
    -- Kick off the call to refresh mods.
    KLEISteamWorkshop:updateWorkshopMods()
end

function mod_manager:enumerateMods()
    util.tclear( self.mods )
    if not KLEISteamWorkshop then
        return
    end

    local steamMods = KLEISteamWorkshop:getWorkshopMods()
    local fsMods = MOAIFileSystem.listDirectories( MOD_FOLDER )
    for i, path in ipairs( fsMods or {} ) do
        local modData = nil
        for i,mod in ipairs(steamMods) do
            if mod.folder == path then
                modData = mod
                break
            end
        end

        if modData == nil then
            -- Didn't exist in Steam mods list, either because that list wasn't updated,
            -- or because this is literally a local Mod that doesn't exist on Steam.
            modData = { folder = path }
        end
        modData.folder = filesystem.pathJoin( MOD_FOLDER, modData.folder )

        self:initMod( modData )
    end
end

function mod_manager:initMod( modData )
    log:write( "initMod - %s", modData.folder )

    local initFile = filesystem.pathJoin( modData.folder, "modinfo.txt" )

    -- Find the specified locale in modinit.
    local fl = io.open( initFile, "r" )
    if fl then
	    for line in fl:lines() do
            local key, value = line:match( "^([_%w]+)[%s]*=[%s]*([_%p%w]+)%s*$" )
            if key and value and modData[ key ] == nil then
                modData[ key ] = value
            end
        end

        if modData.id == nil then
            log:write( "\tMissing 'id' field -- ignoring" )
            return
        end

        table.insert( self.mods, modData )

    else
        log:write( "\tMissing '%s' -- ignoring.", initFile )
    end
end

function mod_manager:mountLanguageMod( id )
    local modData = self:findMod( id )
    if not modData then
        log:write( "Could not mount missing language mod: '%s'", tostring(id) )
    elseif not modData.locale then
        log:write( "Could not mount non-language mod: '%s'", tostring(id) )
    elseif not modData.poFile then
        log:write( "Could not language mod without specified 'poFile': '%s'", tostring(id) )
    else
        log:write( "Mounting language mod: %s ['data-locale' -> '%s']", modData.locale, modData.folder )    
        local oldMount = filesystem.mountVirtualDirectory( "data_locale", modData.folder )
        local loc_translator = include( "loc_translator" )
        if not loc_translator.translateStringTable( "STRINGS", STRINGS, modData.poFile, modData.locale ) then
            filesystem.mountVirtualDirectory( "data_locale", oldMount )
        end
    end
end

function mod_manager:findMod( id )
    for i, modData in ipairs(self.mods) do
        if modData.id == id then
            return modData
        end
    end
end

function mod_manager:getLanguageMods()
    local t = {}
    for i, modData in ipairs(self.mods) do
        if modData.locale then
            table.insert( t, { name = modData.locale, id = modData.id } )
        end
    end
    return t
end

return mod_manager
