local mathutil = include( "modules/mathutil" )
local array = include( "modules/array" )
local util = include( "client_util" )
local simdefs = include("sim/simdefs")
local simquery = include("sim/simquery")
local cdefs = include( "client_defs" )
local mainframe = include( "sim/mainframe" )
local modifiers = include( "sim/modifiers" )
local mission_util = include( "sim/missions/mission_util" )
local serverdefs = include("modules/serverdefs")

local SEED_DISCOUNT = 4
-------------------------------------------------------------------------------
-- These are PC mainframe abilities.  They are owned and executed by the player.
local DEFAULT_ABILITY =
{
	name = nil, -- Override
	desc = nil, -- Override
	cpu_cost = 0, -- Override
	credit_cost = 0, -- Override
	program_strength = 0, --Override 
	break_firewalls = 0, --Override
	equip_program = false, 
	program = true, 

	getCpuCost = function( self )
		local cpuCost = self.cpu_cost
		local zeroCpu = false

		if cpuCost == 0 then 
			zeroCpu = true 
		end

        if self._sim then
            -- Player cost modifier is min-capped at 1, but this doesn't apply to SEED.
            if zeroCpu then 
            	cpuCost = math.max(0, cpuCost + (self._sim:getPC():getTraits().program_cost_modifier or 0))
            else 
            	cpuCost = math.max(1, cpuCost + (self._sim:getPC():getTraits().program_cost_modifier or 0))
            end 

            if (self._sim:getTags().nextProgFree or 0) > 0 then
        	    cpuCost = cpuCost - SEED_DISCOUNT
            end
        end

   		return math.max(0, cpuCost)
	end,

	onTooltip = function( self, screen, sim, player )
		local tooltip = util.tooltip( screen )
		local section = tooltip:addSection()

		if self.toggle_program and self:getCpuCost() > 0 then 
			section:addLine( "<ttheader>"..self.name.."</>", string.format( STRINGS.PROGRAMS.POWER_PER_TURN, self:getCpuCost() ))
		elseif self:getCpuCost() > 0 then 
			section:addLine( "<ttheader>"..self.name.."</>", string.format( STRINGS.PROGRAMS.POWER, self:getCpuCost() ))
		else
			section:addLine( "<ttheader>"..self.name.."</>", string.format( "" ))
		end 
		
		if self.equipped then
			section:addLine( STRINGS.PROGRAMS.EQUIPPED, string.format( "" ))
		end

   		section:addAbility( self.shortdesc, self.desc, "gui/icons/action_icons/Action_icon_Small/icon-item_shoot_small.png" )

        if sim then
		    local canUse, reason = self:canUseAbility( sim, player )
		    if not canUse and reason then
			    section:addRequirement( reason )
		    end
        end

		return tooltip
	end,

	canUseAbility = function( self, sim, abilityOwner, targetUnit )
		local player = sim:getCurrentPlayer()
		if player == nil or player ~= abilityOwner then
			return false
		end

		if (self.cooldown or 0) > 0 then 
			return false, STRINGS.UI.REASON.EQUIPPED_ON_COOLDOWN
		end

		if player:getCpus() < self:getCpuCost() then 
			return false, STRINGS.UI.REASON.NOT_ENOUGH_PWR
		end

		if player:getCredits() < self.credit_cost then
			return false, STRINGS.UI.REASON.NOT_ENOUGH_CREDITS
		end

		if sim:getMainframeLockout() then 
			return false, STRINGS.UI.REASON.INCOGNITA_LOCKED_DOWN
		end 

		for i, x in ipairs(player:getAbilities()) do 
			if x == self then 
				for j, y in ipairs(player:getLockedAbilities()) do 
					if y == i then 
						return false, STRINGS.ABILITIES.TEMPORARY_LOSS 
					end 
				end 
			end 
		end 

		return true 
	end,

   	setEquipped = function( self, isEquipped )
		self.equipped = isEquipped
	end,

    useCPUs = function( self, sim )
        local cdefs = include("client_defs")
        local cpus = self:getCpuCost()
        if (sim:getTags().nextProgFree or 0) > 0 then
            -- Find a seed program with no cooldown.
            for i, ability in ipairs(sim:getCurrentPlayer():getAbilities()) do
                if ability:getID() == "seed" and (ability.cooldown or 0) <= 0 then
                    ability:executeAbility( sim )
                    break
                end
            end
        end
        
        sim:getCurrentPlayer():addCPUs( -cpus, sim )
    end,

	executeAbility = function( self, sim, targetUnit )
        self:useCPUs( sim )

        if targetUnit and (self.break_firewalls or 0) > 0 then
            mainframe.breakIce( sim, targetUnit, self.break_firewalls )
        end
        self.cooldown = self.maxCooldown
	end,		
}


local WRENCH_TEMPLATE = util.extend( DEFAULT_ABILITY )
{
	icon = "gui/icons/programs_icons/icon-program-wrench.png",
	icon_100 = "gui/icons/programs_icons/store_icons/StorePrograms_wrench.png",
	cpu_cost = 2,
	break_firewalls = 0, 
	equip_program = true,
	equipped = false, 
	wrench = true, 
}

local PARASITE_TEMPLATE = util.extend( DEFAULT_ABILITY )
{
	name = STRINGS.PROGRAMS.PARASITE.NAME,
	desc = STRINGS.PROGRAMS.PARASITE.DESC,
	huddesc = STRINGS.PROGRAMS.PARASITE.HUD_DESC,
	shortdesc = STRINGS.PROGRAMS.PARASITE.SHORT_DESC,

	icon = "gui/icons/programs_icons/icon-program-parasite.png",
	icon_100 = "gui/icons/programs_icons/store_icons/StorePrograms_parasite.png",
	cpu_cost = 0,
    break_firewalls = 0,
    parasite_strength = 1,
	equip_program = true, 
	equipped = false, 
	firewallDisplay = "X",
	parasite_hosts = {},

	getCpuCost = function( self )
        self.cpu_cost = #self.parasite_hosts
        return DEFAULT_ABILITY.getCpuCost( self )
	end,

	onSpawnAbility = function( self, sim )
		sim:addTrigger( simdefs.TRG_START_TURN, self )		
	end, 

	onTrigger = function( self, sim, evType, evData )
		if evType == simdefs.TRG_START_TURN and evData:isPC() then

			if self.cooldown and self.cooldown > 0 then 
				self.cooldown = self.cooldown - 1
			end

			local parasiteCount = 0
            local hosts = util.tdupe(self.parasite_hosts)
            for i, hostID in ipairs(hosts) do
                local hostUnit = sim:getUnit( hostID )
                if hostUnit and hostUnit:getTraits().parasite and hostUnit:getTraits().mainframe_ice > 0 then
					mainframe.breakIce( sim, hostUnit, self.parasite_strength )
					parasiteCount = parasiteCount + 1
                end
            end
			if parasiteCount > 0 then
				sim:dispatchEvent( simdefs.EV_MAINFRAME_PARASITE )
			end
		end
	end,


	canUseAbility = function( self, sim, abilityOwner, targetUnit )
		if targetUnit and targetUnit:getTraits().parasite then
			return false, STRINGS.PROGRAMS.PARASITE.ALREADY_HOSTED
		end

		return DEFAULT_ABILITY.canUseAbility( self, sim, abilityOwner, targetUnit )
	end,

	executeAbility = function( self, sim, targetUnit )
        if not targetUnit:getTraits().parasite then
			sim:dispatchEvent( simdefs.EV_PLAY_SOUND, simdefs.SOUND_HOST_PARASITE.path )
			self.cooldown = self.maxCooldown
            self:useCPUs( sim )
			targetUnit:getTraits().parasite = true 
			if self.parasiteV2 then
				targetUnit:getTraits().parasiteV2 = true 
			end
			sim:dispatchEvent( simdefs.EV_PLAY_SOUND, "SpySociety/Actions/mainframe_parasite_installed")
			table.insert(self.parasite_hosts,targetUnit:getID())
        end
	end,		
}

local mainframe_abilities =
{
	remoteprocessor = util.extend( DEFAULT_ABILITY ) 
	{
		name = STRINGS.PROGRAMS.POWER_DRIP.NAME,
		desc = STRINGS.PROGRAMS.POWER_DRIP.DESC,
		shortdesc = STRINGS.PROGRAMS.POWER_DRIP.SHORT_DESC,
		huddesc = STRINGS.PROGRAMS.POWER_DRIP.HUD_DESC,
		icon = "gui/icons/programs_icons/icon-program-powerdrip.png",
		icon_100 = "gui/icons/programs_icons/store_icons/StorePrograms_drip.png",
		credit_cost = 0, 
		value = 300,

		passive = true,
		
		executeAbility = function( self, sim )
			local player = sim:getCurrentPlayer()			
			if not player:isNPC() then
				--jcheng: no longer tell them about power drip
				--[[
				if sim:getParams().missionCount == 0 and sim:getTurnCount() == 2 and not sim:getTags().isTutorial then
                    local dialogParams =
                    {
                        STRINGS.PROGRAMS.POWER_DRIP.MODAL_1,
                        STRINGS.PROGRAMS.POWER_DRIP.MODAL_2,
                        STRINGS.PROGRAMS.POWER_DRIP.MODAL_3,
                        "gui/icons/programs_icons/store_icons/StorePrograms_drip.png"
                    }
					sim:dispatchEvent( simdefs.EV_SHOW_DIALOG, { dialog = "programDialog", dialogParams = dialogParams } )
				end	
				]]

				sim:dispatchEvent( simdefs.EV_PLAY_SOUND, cdefs.SOUND_HUD_MAINFRAME_PROGRAM_AUTO_RUN )
				sim:dispatchEvent( simdefs.EV_SHOW_WARNING, {txt=STRINGS.PROGRAMS.POWER_DRIP.WARNING, color=cdefs.COLOR_PLAYER_WARNING, sound = "SpySociety/Actions/mainframe_gainCPU",icon=self.icon } )
				player:addCPUs( 1 )
			end
		end,

		canUseAbility = function( self, sim )
			return false 	
		end,

		onSpawnAbility = function( self, sim )
			sim:addTrigger( simdefs.TRG_START_TURN, self )		
		end,


		onTrigger = function( self, sim, evType, evData )
			if evType == simdefs.TRG_START_TURN then
				self:executeAbility(sim)	
			end
		end,
	},

	emergency_drip = util.extend( DEFAULT_ABILITY ) 
	{
		name = STRINGS.PROGRAMS.EMERGENCY_DRIP.NAME,
		desc = STRINGS.PROGRAMS.EMERGENCY_DRIP.DESC,
		shortdesc = STRINGS.PROGRAMS.EMERGENCY_DRIP.SHORT_DESC,
		huddesc = STRINGS.PROGRAMS.EMERGENCY_DRIP.HUD_DESC,
		icon = "gui/icons/programs_icons/Program0016.png",
		icon_100 = "gui/icons/programs_icons/store_icons/StorePrograms_0016.png",
		cpu_cost = nil,
		credit_cost = 0, 
		value = 300,
        MIN_CPU = 4,
        cooldown = 0,
        maxCooldown = 3, 

        passive = true,

		executeAbility = function( self, sim )
			local player = sim:getCurrentPlayer()			
			if not player:isNPC() and player:getCpus() < self.MIN_CPU and self.cooldown <= 0 then
				sim:dispatchEvent( simdefs.EV_SHOW_WARNING, {txt=string.format( STRINGS.PROGRAMS.EMERGENCY_DRIP.WARNING, 5 ), color=cdefs.COLOR_PLAYER_WARNING, sound = "SpySociety/Actions/mainframe_gainCPU",icon=self.icon } )
				player:addCPUs( 5 )
				self.cooldown = self.maxCooldown
			end
		end,

		canUseAbility = function( self, sim )
			return false 	
		end,

		onSpawnAbility = function( self, sim )
			sim:addTrigger( simdefs.TRG_START_TURN, self )		
		end,

		onTrigger = function( self, sim, evType, evData )
			if evType == simdefs.TRG_START_TURN and evData:isPC() then
				if self.cooldown > 0 then 
					self.cooldown = self.cooldown - 1
				end 

				if self.cooldown <= 0 then 
					self:executeAbility( sim )
				end 
			end
		end,
	},

	lockpick_1 = util.extend( DEFAULT_ABILITY )
	{

		name = STRINGS.PROGRAMS.LOCKPICK.NAME,
		desc = STRINGS.PROGRAMS.LOCKPICK.DESC,
		huddesc = STRINGS.PROGRAMS.LOCKPICK.HUD_DESC,
		shortdesc = STRINGS.PROGRAMS.LOCKPICK.SHORT_DESC,
		tipdesc = STRINGS.PROGRAMS.LOCKPICK.TIP_DESC,

		icon = "gui/icons/programs_icons/icon-program-lockpick.png",
		icon_100 = "gui/icons/programs_icons/store_icons/StorePrograms_lockpick.png",
		cpu_cost = 2,
		break_firewalls = 1, 
		equip_program = true, 
		equipped = false, 
		value = 300,
	},

	lockpick_2 = util.extend( DEFAULT_ABILITY )
	{

		name = STRINGS.PROGRAMS.LOCKPICK_2.NAME,
		desc = STRINGS.PROGRAMS.LOCKPICK_2.DESC,
		huddesc = STRINGS.PROGRAMS.LOCKPICK_2.HUD_DESC,
		shortdesc = STRINGS.PROGRAMS.LOCKPICK_2.SHORT_DESC,
		tipdesc = STRINGS.PROGRAMS.LOCKPICK_2.TIP_DESC,

		icon = "gui/icons/programs_icons/icon-program-lockpick_2.png",
		icon_100 = "gui/icons/programs_icons/store_icons/StorePrograms_lockpick_2.png",
		cpu_cost = 3,
		break_firewalls = 2, 
		equip_program = true, 
		noexecute = true, 
		equipped = false, 
		value = 700,
	},

	wrench_2 = util.extend( WRENCH_TEMPLATE )
	{
		name = STRINGS.PROGRAMS.WRENCH_2.NAME,
		desc = STRINGS.PROGRAMS.WRENCH_2.DESC,
		huddesc = STRINGS.PROGRAMS.WRENCH_2.HUD_DESC,
		shortdesc = STRINGS.PROGRAMS.WRENCH_2.SHORT_DESC,
		tipdesc = STRINGS.PROGRAMS.WRENCH_2.TIP_DESC,
		break_firewalls = 2,
		value = 600,
	},

	wrench_3 = util.extend( WRENCH_TEMPLATE )
	{
		name = STRINGS.PROGRAMS.WRENCH_3.NAME,
		desc = STRINGS.PROGRAMS.WRENCH_3.DESC,
		huddesc = STRINGS.PROGRAMS.WRENCH_3.HUD_DESC,
		shortdesc = STRINGS.PROGRAMS.WRENCH_3.SHORT_DESC,
		tipdesc = STRINGS.PROGRAMS.WRENCH_3.TIP_DESC,
		break_firewalls = 3,
		cpu_cost = 3,
		value = 600,
	},

	wrench_4 = util.extend( WRENCH_TEMPLATE )
	{
		name = STRINGS.PROGRAMS.WRENCH_4.NAME,
		desc = STRINGS.PROGRAMS.WRENCH_4.DESC,
		huddesc = STRINGS.PROGRAMS.WRENCH_4.HUD_DESC,
		shortdesc = STRINGS.PROGRAMS.WRENCH_4.SHORT_DESC,
		tipdesc = STRINGS.PROGRAMS.WRENCH_4.TIP_DESC,
		break_firewalls = 4,
		cpu_cost = 4,
		value = 600,
	},

	wrench_5 = util.extend( WRENCH_TEMPLATE )
	{
		name = STRINGS.PROGRAMS.WRENCH_5.NAME,
		desc = STRINGS.PROGRAMS.WRENCH_5.DESC,
		huddesc = STRINGS.PROGRAMS.WRENCH_5.HUD_DESC,
		shortdesc = STRINGS.PROGRAMS.WRENCH_5.SHORT_DESC,
		tipdesc = STRINGS.PROGRAMS.WRENCH_5.TIP_DESC,
		break_firewalls = 5,
		cpu_cost = 4,
		value = 600,
	},

	hammer = util.extend( DEFAULT_ABILITY )
	{

		name = STRINGS.PROGRAMS.HAMMER.NAME,
		desc = STRINGS.PROGRAMS.HAMMER.DESC,
		huddesc = STRINGS.PROGRAMS.HAMMER.HUD_DESC,
		shortdesc = STRINGS.PROGRAMS.HAMMER.SHORT_DESC,
		tipdesc = STRINGS.PROGRAMS.HAMMER.TIP_DESC,

		icon = "gui/icons/programs_icons/icon-program-hammer.png",
		icon_100 = "gui/icons/programs_icons/store_icons/StorePrograms_hammer.png",
		cpu_cost = 5,
		break_firewalls = 5, 
		equip_program = true, 
		equipped = false, 
		cooldown = 0,
		maxCooldown = 3,
		value = 600,

		onSpawnAbility = function( self, sim )
			sim:addTrigger( simdefs.TRG_START_TURN, self )		
		end,

		onTrigger = function( self, sim, evType, evData )
			if evType == simdefs.TRG_START_TURN and evData:isPC() then
				if self.cooldown > 0 then 
					self.cooldown = self.cooldown - 1
				end 
			end
		end,
	},

	dagger = util.extend( DEFAULT_ABILITY )
	{
		name = STRINGS.PROGRAMS.DAGGER.NAME,
		desc = STRINGS.PROGRAMS.DAGGER.DESC,
		huddesc = STRINGS.PROGRAMS.DAGGER.HUD_DESC,
		shortdesc = STRINGS.PROGRAMS.DAGGER.SHORT_DESC,
		tipdesc = STRINGS.PROGRAMS.DAGGER.TIP_DESC,

		icon = "gui/icons/programs_icons/icon-program-dagger.png",
		icon_100 = "gui/icons/programs_icons/store_icons/StorePrograms_dagger.png",
		cpu_cost = 1,
		break_firewalls = 3, 
		equip_program = true, 
		equipped = false, 
		cooldown = 0,
		maxCooldown = 5,
		value = 450,

		onSpawnAbility = function( self, sim )
			sim:addTrigger( simdefs.TRG_START_TURN, self )		
		end,

		onTrigger = function( self, sim, evType, evData )
			if evType == simdefs.TRG_START_TURN and evData:isPC() then
				if self.cooldown > 0 then 
					self.cooldown = self.cooldown - 1
				end 
			end
		end,
	},

	brimstone = util.extend( DEFAULT_ABILITY )
	{
		name = STRINGS.PROGRAMS.BRIMSTONE.NAME,
		desc = STRINGS.PROGRAMS.BRIMSTONE.DESC,
		huddesc = STRINGS.PROGRAMS.BRIMSTONE.HUD_DESC,
		shortdesc = STRINGS.PROGRAMS.BRIMSTONE.SHORT_DESC,
		tipdesc = STRINGS.PROGRAMS.BRIMSTONE.TIP_DESC,

		icon = "gui/icons/programs_icons/Program0025.png",
		icon_100 = "gui/icons/programs_icons/store_icons/StorePrograms_brimstone.png",
		cpu_cost = 3,
		break_firewalls = 1, 
		equip_program = true, 
		equipped = false, 
		value = 450,
		daemonReversalAdd = 10,
		lockedText = STRINGS.UI.TEAM_SELECT.UNLOCK_CENTRAL_MONSTER,
	},

	dagger_2 = util.extend( DEFAULT_ABILITY )
	{


		name = STRINGS.PROGRAMS.DAGGER_2.NAME,
		desc = STRINGS.PROGRAMS.DAGGER_2.DESC,
		huddesc = STRINGS.PROGRAMS.DAGGER_2.HUD_DESC,
		shortdesc = STRINGS.PROGRAMS.DAGGER_2.SHORT_DESC,
		tipdesc = STRINGS.PROGRAMS.DAGGER_2.TIP_DESC,

		icon = "gui/icons/programs_icons/icon-program-dagger_2.png",
		icon_100 = "gui/icons/programs_icons/store_icons/StorePrograms_dagger_2.png",
		cpu_cost = 3,
		break_firewalls = 5, 
		equip_program = true, 
		noexecute = true, 
		equipped = false, 
		cooldown = 0,
		maxCooldown = 4,
		value = 650,

		onSpawnAbility = function( self, sim )
			sim:addTrigger( simdefs.TRG_START_TURN, self )		
		end,

		onTrigger = function( self, sim, evType, evData )
			if evType == simdefs.TRG_START_TURN and evData:isPC() then
				if self.cooldown > 0 then 
					self.cooldown = self.cooldown - 1
				end 
			end
		end,
	},

	parasite = util.extend( PARASITE_TEMPLATE )
	{
		value = 400,
		tipdesc = STRINGS.PROGRAMS.PARASITE.TIP_DESC,
    },

	parasite_2 = util.extend( PARASITE_TEMPLATE )
	{

		name = STRINGS.PROGRAMS.PARASITE_2.NAME,
		desc = STRINGS.PROGRAMS.PARASITE_2.DESC,
		huddesc = STRINGS.PROGRAMS.PARASITE_2.HUD_DESC,
		shortdesc = STRINGS.PROGRAMS.PARASITE_2.SHORT_DESC,
		tipdesc = STRINGS.PROGRAMS.PARASITE_2.TIP_DESC,

		icon = "gui/icons/programs_icons/icon-program-parasite_2.png",
		icon_100 = "gui/icons/programs_icons/store_icons/StorePrograms_parasite_2.png",
		cpu_cost = 2,
        parasite_strength = 2,
        value = 500,
        parasiteV2 = true,

	    getCpuCost = function( self )
            self.cpu_cost = #self.parasite_hosts + 2
            return DEFAULT_ABILITY.getCpuCost( self )
	    end,
	},

	sniffer = util.extend( DEFAULT_ABILITY )
	{
		name = STRINGS.PROGRAMS.SNIFFER.NAME,
		desc = STRINGS.PROGRAMS.SNIFFER.DESC,
		huddesc = STRINGS.PROGRAMS.SNIFFER.HUD_DESC,
		shortdesc = STRINGS.PROGRAMS.SNIFFER.SHORT_DESC,
		tipdesc = STRINGS.PROGRAMS.SNIFFER.TIP_DESC,

		icon = "gui/icons/programs_icons/icon-program-sniffer.png",
		icon_100 = "gui/icons/programs_icons/store_icons/StorePrograms_sniffer.png",
		cpu_cost = 0,
		break_firewalls = 0, 
		equip_program = true, 
		equipped = false,
		sniffer = true,  
		cooldown = 0,
		maxCooldown = 3,
		value = 350,

		onSpawnAbility = function( self, sim )
			sim:addTrigger( simdefs.TRG_START_TURN, self )		
		end,
		
		onTrigger = function( self, sim, evType, evData )
			if evType == simdefs.TRG_START_TURN and evData:isPC() then
				if self.cooldown > 0 then 
					self.cooldown = self.cooldown - 1
				end 
			end
		end,

		executeAbility = function( self, sim, targetUnit )
			targetUnit:getTraits().daemon_sniffed = true 
			self.cooldown = self.maxCooldown
            self:useCPUs( sim )
			sim:dispatchEvent( simdefs.EV_PLAY_SOUND, simdefs.SOUND_DAEMON_REVEAL.path )
		end,
	},

	hunter = util.extend( DEFAULT_ABILITY )
	{
		name = STRINGS.PROGRAMS.HUNTER.NAME,
		desc = STRINGS.PROGRAMS.HUNTER.DESC,
		huddesc = STRINGS.PROGRAMS.HUNTER.HUD_DESC,
		shortdesc = STRINGS.PROGRAMS.HUNTER.SHORT_DESC,
		tipdesc = STRINGS.PROGRAMS.HUNTER.TIP_DESC,

		icon = "gui/icons/programs_icons/Programs0013.png",
		icon_100 = "gui/icons/programs_icons/store_icons/StorePrograms_program13.png",
		cpu_cost = 5,
		break_firewalls = 0, 
		equip_program = true, 
		equipped = false,
		daemon_killer = true,   
		cooldown = 0,
		maxCooldown = 3,
		value = 550,

		onSpawnAbility = function( self, sim )
			sim:addTrigger( simdefs.TRG_START_TURN, self )		
		end,
		
		onTrigger = function( self, sim, evType, evData )
			if evType == simdefs.TRG_START_TURN and evData:isPC() then
				if self.cooldown > 0 then 
					self.cooldown = self.cooldown - 1
				end 
			end
		end,

		executeAbility = function( self, sim, targetUnit )

			if targetUnit:getTraits().mainframe_program == nil then
				sim:dispatchEvent( simdefs.EV_PLAY_SOUND, simdefs.SOUND_HUD_INCIDENT_NEGATIVE.path )
				mission_util.showDialog( sim, STRINGS.UI.DIALOGS.NO_DAEMON_TITLE, STRINGS.UI.DIALOGS.NO_DAEMON_BODY )
			else
			
				targetUnit:getTraits().mainframe_program = nil
				sim:dispatchEvent( simdefs.EV_KILL_DAEMON, {unit = targetUnit})
				
				if targetUnit:getTraits().daemonHost then
					sim:getUnit(targetUnit:getTraits().daemonHost):killUnit(sim)
					targetUnit:getTraits().daemonHost =nil
				end

				self.cooldown = self.maxCooldown
                self:useCPUs( sim )

				sim:dispatchEvent( simdefs.EV_PLAY_SOUND, simdefs.SOUND_DAEMON_REVEAL.path )
			end
		end,
	},

	mainframePing = util.extend( DEFAULT_ABILITY )
	{
		name = STRINGS.PROGRAMS.PING.NAME,
		desc = STRINGS.PROGRAMS.PING.DESC,
		huddesc = STRINGS.PROGRAMS.PING.HUD_DESC,
		shortdesc = STRINGS.PROGRAMS.PING.SHORT_DESC,
		tipdesc = STRINGS.PROGRAMS.PING.TIP_DESC,


		icon = "gui/icons/programs_icons/icon-program-ping.png",
		icon_100 = "gui/icons/programs_icons/store_icons/StorePrograms_ping.png",
		cpu_cost = 1,
		credit_cost = 0,
		cooldown = 0,
		maxCooldown = 4,
		value = 450,

		onSpawnAbility = function( self, sim )
			sim:addTrigger( simdefs.TRG_START_TURN, self )		
		end,

		onTrigger = function( self, sim, evType, evData )
			if evType == simdefs.TRG_START_TURN and evData:isPC() then
				if self.cooldown > 0 then 
					self.cooldown = self.cooldown - 1
				end 
			end
		end,

		acquireTargets = function( self, targets, game, sim, unit )
			return targets.simpleAreaTarget( game, simdefs.SOUND_MAINFRAME_PING.range, sim )
		end, 

		executeAbility = function( self, sim, unit, userUnit, targetCell )
			local player = sim:getCurrentPlayer()
			local cellx, celly = unpack(targetCell)

			self.cooldown = self.maxCooldown

			sim:dispatchEvent( simdefs.EV_SCANRING_VIS, { x= cellx,y= celly, range= simdefs.SOUND_MAINFRAME_PING.range } )		

			sim:emitSound( simdefs.SOUND_MAINFRAME_PING, cellx, celly, nil, {{ x = cellx, y = celly }} )
			sim:processReactions()
			-- Always want to hear the sound played
			sim:dispatchEvent( simdefs.EV_PLAY_SOUND, simdefs.SOUND_MAINFRAME_PING.path )
            self:useCPUs( sim )
			player:addCredits( -self.credit_cost )
		end,
	},

	fusion = util.extend( DEFAULT_ABILITY )
	{

		name = STRINGS.PROGRAMS.FUSION.NAME,
		desc = STRINGS.PROGRAMS.FUSION.DESC,
		huddesc = STRINGS.PROGRAMS.FUSION.HUD_DESC,
		shortdesc = STRINGS.PROGRAMS.FUSION.SHORT_DESC,
		tipdesc = STRINGS.PROGRAMS.FUSION.TIP_DESC,

		icon = "gui/icons/programs_icons/icon-program-fusion.png",
		icon_100 = "gui/icons/programs_icons/store_icons/StorePrograms_fusion.png",
		cpu_cost = 5,
		credit_cost = 0,
		cooldown = 0,
		maxCooldown = 4,
		value = 300,

		onSpawnAbility = function( self, sim )
			sim:addTrigger( simdefs.TRG_START_TURN, self )		
		end,

		onTrigger = function( self, sim, evType, evData )
			if evType == simdefs.TRG_START_TURN and evData:isPC() then
				if self.cooldown > 0 then 
					self.cooldown = self.cooldown - 1
					evData:addCPUs( 3 )
					sim:dispatchEvent( simdefs.EV_PLAY_SOUND, cdefs.SOUND_HUD_MAINFRAME_PROGRAM_AUTO_RUN )
					sim:dispatchEvent( simdefs.EV_SHOW_WARNING, {txt=STRINGS.PROGRAMS.FUSION.WARNING, color=cdefs.COLOR_PLAYER_WARNING, sound = "SpySociety/Actions/mainframe_gainCPU",icon=self.icon } )
				end 
			end
		end,

		executeAbility = function( self, sim, unit, userUnit, targetCell )
			local player = sim:getCurrentPlayer()
			self.cooldown = self.maxCooldown
			sim:dispatchEvent( simdefs.EV_PLAY_SOUND, "SpySociety/Actions/mainframe_run_fusion" )
            self:useCPUs( sim )
		end,
	},

	seed = util.extend( DEFAULT_ABILITY )
	{

		name = STRINGS.PROGRAMS.SEED.NAME,
		desc = STRINGS.PROGRAMS.SEED.DESC,
		huddesc = STRINGS.PROGRAMS.SEED.HUD_DESC,
		shortdesc = STRINGS.PROGRAMS.SEED.SHORT_DESC,
		tipdesc = STRINGS.PROGRAMS.SEED.TIP_DESC,

		icon = "gui/icons/programs_icons/icon-program-seed.png",
		icon_100 = "gui/icons/programs_icons/store_icons/StorePrograms_seed.png",
		cpu_cost = 0,
		credit_cost = 0,
		cooldown = 0,
		maxCooldown = 1,
		passive = true,
		value = 350,

		canUseAbility = function( self, sim )
			return false 
		end,

		onSpawnAbility = function( self, sim )
			sim:addTrigger( simdefs.TRG_START_TURN, self )		
		end,

        executeAbility = function( self, sim )
            self.cooldown = self.maxCooldown

            sim:dispatchEvent( simdefs.EV_PLAY_SOUND, cdefs.SOUND_HUD_MAINFRAME_PROGRAM_AUTO_RUN )
		    sim:dispatchEvent( simdefs.EV_SHOW_WARNING, {txt=STRINGS.PROGRAMS.SEED.WARNING, color=cdefs.COLOR_PLAYER_WARNING, sound = "SpySociety/Actions/mainframe_gainCPU",icon=self.icon } )

            sim:getTags().nextProgFree = (sim:getTags().nextProgFree or 0) - 1
            if sim:getTags().nextProgFree <= 0 then
                sim:getTags().nextProgFree = nil
            end
        end,

		onTrigger = function( self, sim, evType, evData )
			if evType == simdefs.TRG_START_TURN and evData:isPC() then
				if self.cooldown > 0 then 
					self.cooldown = self.cooldown - 1
				end 
                if self.cooldown == 0 then
        			sim:getTags().nextProgFree = (sim:getTags().nextProgFree or 0) + 1
                end
			end
		end,
	},


	faust = util.extend( DEFAULT_ABILITY )
	{
		name = STRINGS.PROGRAMS.FAUST.NAME,
		desc = STRINGS.PROGRAMS.FAUST.DESC,
		huddesc = STRINGS.PROGRAMS.FAUST.HUD_DESC,
		shortdesc = STRINGS.PROGRAMS.FAUST.SHORT_DESC,
		tipdesc = STRINGS.PROGRAMS.FAUST.TIP_DESC,
		lockedText = STRINGS.UI.TEAM_SELECT.UNLOCK_CENTRAL_MONSTER,

		icon = "gui/icons/programs_icons/icon-program-faust.png",
		icon_100 = "gui/icons/programs_icons/store_icons/StorePrograms_faust.png",
		credit_cost = 0, 
		value = 300,
		daemonChance = 20,
		lastTurnDaemon = false,

		passive = true,
		
		executeAbility = function( self, sim )
			--don't execute on first turn
			if sim:getTurnCount() == 1 then
				return
			end

			local player = sim:getCurrentPlayer()			
			if not player:isNPC() then
				sim:dispatchEvent( simdefs.EV_PLAY_SOUND, cdefs.SOUND_HUD_MAINFRAME_PROGRAM_AUTO_RUN )
				sim:dispatchEvent( simdefs.EV_SHOW_WARNING, {txt=STRINGS.PROGRAMS.FAUST.WARNING, color=cdefs.COLOR_PLAYER_WARNING, sound = "SpySociety/Actions/mainframe_gainCPU",icon=self.icon } )
				player:addCPUs( 2 )
				local rand = sim:nextRand( 0, 100 )
				if rand <= self.daemonChance then 
					local programList = serverdefs.PROGRAM_LIST
					local daemon = programList[sim:nextRand(1, #programList)]
					sim:getNPC():addMainframeAbility( sim, daemon )
					self.lastTurnDaemon = true 
				end 
			end
		end,

		canUseAbility = function( self, sim )
			return false 	
		end,

		onSpawnAbility = function( self, sim )
			sim:addTrigger( simdefs.TRG_START_TURN, self )		
		end,


		onTrigger = function( self, sim, evType, evData )
			if evType == simdefs.TRG_START_TURN then
				self:executeAbility(sim)	
			end
		end,
	},

	love = util.extend( DEFAULT_ABILITY )
	{

		name = STRINGS.PROGRAMS.LOVE.NAME,
		desc = STRINGS.PROGRAMS.LOVE.DESC,
		huddesc = STRINGS.PROGRAMS.LOVE.HUD_DESC,
		shortdesc = STRINGS.PROGRAMS.LOVE.SHORT_DESC,
		tipdesc = STRINGS.PROGRAMS.LOVE.TIP_DESC,

		icon = "gui/icons/programs_icons/Program0024.png",
		icon_100 = "gui/icons/programs_icons/store_icons/StorePrograms_love.png",
		cpu_cost = 5,
		credit_cost = 0,
		cooldown = 0,
		maxCooldown = 3,
		program_strength = 2, 
		--color = util.color( 39/255, 215/255, 39/255, 1 ),

		onSpawnAbility = function( self, sim )
			sim:addTrigger( simdefs.TRG_START_TURN, self )		
		end,

		onTrigger = function( self, sim, evType, evData )
			if evType == simdefs.TRG_START_TURN and evData:isPC() then
				if self.cooldown > 0 then 
					self.cooldown = self.cooldown - 1
				end 
			end
		end,
        
        getTargetUnits = function( self, sim )
            local units = {}
            local x0, y0
            for unitID, unit in pairs( sim:getAllUnits() ) do
                local x, y = unit:getLocation()
                if x and y and sim:getCurrentPlayer():getLastKnownCell( sim, x, y ) ~= nil then
                    if unit:getTraits().central then
                        x0, y0 = x, y
                    elseif mainframe.canBreakIce( sim, unit, self ) then
                        table.insert( units, unit )
                    end
                end
            end

            if not x0 then
                return -- No ground zero.
            end

            local range = -math.huge
            if x0 and y0 then
                for i, unit in ipairs(units) do
                    range = math.max( range, mathutil.dist2d( x0, y0, unit:getLocation() ))
                end
            end
            return units, x0, y0, math.ceil( range )
        end,

		executeAbility = function( self, sim, unit, userUnit )
			local player = sim:getCurrentPlayer()
            local currentProgram = player:getEquippedProgram()

    		player:equipProgram( sim, self:getID() )

            sim:dispatchEvent( simdefs.EV_PLAY_SOUND, "SpySociety/Actions/mainframe_turing_use" )

            local targetUnits, x0, y0, range = self:getTargetUnits( sim )
            if not x0 then
                return
            end
            -- For effect, ensure range is at least reasonably large...
            range = math.max( range, 16 )
           	sim:dispatchEvent( simdefs.EV_OVERLOAD_VIZ, {x = x0, y = y0, units = targetUnits, range = range } )		

            local daemonUnits = {}
            for _, unit in ipairs(targetUnits) do
                -- Keep track of daemons that *Would* be invoked, so that we can invoke them only after
                -- everything has been broken.
                local daemonProgram = unit:getTraits().mainframe_program

                unit:getTraits().mainframe_program = nil
                mainframe.breakIce( sim, unit, self.program_strength )
                unit:getTraits().mainframe_program = daemonProgram

                if daemonProgram and unit:getTraits().mainframe_ice <= 0 then
                    table.insert( daemonUnits, unit )
                end
	        end

            self:useCPUs( sim )
    		player:equipProgram( sim, currentProgram and currentProgram:getID() )

            for i, unit in ipairs( daemonUnits ) do
                mainframe.invokeDaemon( sim, unit )
            end
            self.cooldown = self.maxCooldown

            sim:triggerEvent( "used_turing" )            
		end,
	},

	wildfire = util.extend( DEFAULT_ABILITY )
	{

		name = STRINGS.PROGRAMS.WILDFIRE.NAME,
		desc = STRINGS.PROGRAMS.WILDFIRE.DESC,
		huddesc = STRINGS.PROGRAMS.WILDFIRE.HUD_DESC,
		shortdesc = STRINGS.PROGRAMS.WILDFIRE.SHORT_DESC,
		tipdesc = STRINGS.PROGRAMS.WILDFIRE.TIP_DESC,

		icon = "gui/icons/programs_icons/icon-program-wildfire.png",
		icon_100 = "gui/icons/programs_icons/store_icons/StorePrograms_wildfire.png",
		cpu_cost = 0,
		credit_cost = 0,
		cooldown = 0,
		maxCooldown = 1,
		value = 400,

		onSpawnAbility = function( self, sim )
			sim:addTrigger( simdefs.TRG_START_TURN, self )		
		end,

		onTrigger = function( self, sim, evType, evData )
			if evType == simdefs.TRG_START_TURN and evData:isPC() then
				if self.cooldown > 0 then 
					self.cooldown = self.cooldown - 1
				end 
			end
		end,

		executeAbility = function( self, sim, unit, userUnit, targetCell )
			local player = sim:getCurrentPlayer()
			self.cooldown = self.maxCooldown
			sim:dispatchEvent( simdefs.EV_PLAY_SOUND, "SpySociety/Actions/mainframe_run_fusion" )
            self:useCPUs( sim )
			player:addCPUs( 5 )
			sim:dispatchEvent( simdefs.EV_SHOW_WARNING, {txt=STRINGS.PROGRAMS.WILDFIRE.WARNING, color=cdefs.COLOR_PLAYER_WARNING, sound = "SpySociety/Actions/mainframe_gainCPU",icon=self.icon } )
			
			for _, unit in pairs( sim:getAllUnits() ) do			
		    	unit:increaseIce(sim,1)
			end

		end,
	},


	oracle = util.extend( DEFAULT_ABILITY )
	{

		name = STRINGS.PROGRAMS.ORACLE.NAME,
		desc = STRINGS.PROGRAMS.ORACLE.DESC,
		huddesc = STRINGS.PROGRAMS.ORACLE.HUD_DESC,
		shortdesc = STRINGS.PROGRAMS.ORACLE.SHORT_DESC,
		tipdesc = STRINGS.PROGRAMS.ORACLE.TIP_DESC,

		icon = "gui/icons/programs_icons/icon-program-oracle.png",
		icon_100 = "gui/icons/programs_icons/store_icons/StorePrograms_oracle.png",
		cpu_cost = 4,
		credit_cost = 0,
		value = 500,

		canUseAbility = function( self, sim, abilityOwner )
			local player = sim:getCurrentPlayer()
			if player ~= abilityOwner or player == nil then
				return false
			end
			--Are there any uncaptured cameras left? 
			local cameras = false
			for _, unit in pairs( sim:getAllUnits() ) do 	
				if unit:getTraits().mainframe_camera then 
					if unit:getPlayerOwner() ~= player and unit:getTraits().mainframe_status == "active" then 
						cameras = true 
					end
				end
			end

			if not cameras then 
				return false, STRINGS.UI.REASON.NO_CAMERAS
			end

			return DEFAULT_ABILITY.canUseAbility( self, sim, abilityOwner )
		end,

		executeAbility = function( self, sim, unit, userUnit, targetCell )
			local player = sim:getCurrentPlayer()
            self:useCPUs( sim )
			local availableCameras = {}

			for _, unit in pairs( sim:getAllUnits() ) do 
				if unit:getTraits().mainframe_camera then 
					if unit:getPlayerOwner() ~= player and unit:getTraits().mainframe_status == "active" then 
						table.insert( availableCameras, unit )
					end
				end
			end

			local camera = availableCameras[ sim:nextRand( 1, #availableCameras ) ]
            mainframe.breakIce( sim, camera, camera:getTraits().mainframe_ice )

    		sim:dispatchEvent( simdefs.EV_CAM_PAN, { camera:getLocation() } )	
			sim:dispatchEvent( simdefs.EV_PLAY_SOUND, "SpySociety/Actions/mainframe_object_camera" )
		end,
	},

	rapier = util.extend( DEFAULT_ABILITY )
	{
		name = STRINGS.PROGRAMS.RAPIER.NAME,
		desc = STRINGS.PROGRAMS.RAPIER.DESC,
		huddesc = STRINGS.PROGRAMS.RAPIER.HUD_DESC,
		shortdesc = STRINGS.PROGRAMS.RAPIER.SHORT_DESC,
		tipdesc = STRINGS.PROGRAMS.RAPIER.TIP_DESC,

		icon = "gui/icons/programs_icons/Program0017.png",
		icon_100 = "gui/icons/programs_icons/store_icons/StorePrograms_0017.png",
		cpu_cost = 1,
		break_firewalls = 1, 
		equip_program = true, 
		noexecute = true, 
		equipped = false, 
		value = 600,
		current_alarm = 0,

		getCpuCost = function( self )
            local trackerStage = 0
            if self._sim then
        	    trackerStage = self._sim:getTrackerStage( math.min( simdefs.TRACKER_MAXCOUNT, self._sim:getTracker() ))
            end
            self.cpu_cost = 1 + trackerStage

            return DEFAULT_ABILITY.getCpuCost( self )
		end,

		onSpawnAbility = function( self, sim )
            self._sim = sim
		end,
	},

	wings = util.extend( DEFAULT_ABILITY )
	{

		name = STRINGS.PROGRAMS.WINGS.NAME,
		desc = STRINGS.PROGRAMS.WINGS.DESC,
		huddesc = STRINGS.PROGRAMS.WINGS.HUD_DESC,
		shortdesc = STRINGS.PROGRAMS.WINGS.SHORT_DESC,
		tipdesc = STRINGS.PROGRAMS.WINGS.TIP_DESC,

		icon = "gui/icons/programs_icons/icon-program-wings.png",
		icon_100 = "gui/icons/programs_icons/store_icons/StorePrograms_wings.png",
		cpu_cost = 1,
		cooldown = 0,
		maxCooldown = 2,
		value = 500,

		onSpawnAbility = function( self, sim )
			sim:addTrigger( simdefs.TRG_START_TURN, self )		
		end,

		onTrigger = function( self, sim, evType, evData )
			if evType == simdefs.TRG_START_TURN and evData:isPC() then
				if self.cooldown > 0 then 
					self.cooldown = self.cooldown - 1
				end 
			end
		end,

		executeAbility = function( self, sim )
			self.cooldown = self.maxCooldown
			local player = sim:getCurrentPlayer()
            self:useCPUs( sim )

            sim:dispatchEvent( simdefs.EV_PLAY_SOUND, "SpySociety/Actions/mainframe_wings_activate")

			for _, unit in pairs(sim:getAllUnits()) do
				if unit:getPlayerOwner() == player and simquery.isAgent( unit ) then
					unit:getTraits().mp = unit:getTraits().mp + 2
					local x0,y0 = unit:getLocation()
					sim:dispatchEvent( simdefs.EV_UNIT_FLOAT_TXT, {txt=util.sformat(STRINGS.FORMATS.PLUS_AP,2),x=x0,y=y0,color={r=1,g=1,b=1,a=1}} )		

					sim:dispatchEvent( simdefs.EV_UNIT_REFRESH, { unit = unit } )

					local x1, y1 = unit:getLocation()
					sim:dispatchEvent( simdefs.EV_GAIN_AP, { unit = unit } )
					sim:dispatchEvent( simdefs.EV_UNIT_FLOAT_TXT, {txt = STRINGS.PROGRAMS.WINGS.NAME, x = x1, y = y1,color={r=255/255,g=255/255,b=51/255,a=1}} )	
				end
			end
		end,
	},

	shade = util.extend( DEFAULT_ABILITY )
	{


		name = STRINGS.PROGRAMS.SHADE.NAME,
		desc = STRINGS.PROGRAMS.SHADE.DESC,
		huddesc = STRINGS.PROGRAMS.SHADE.HUD_DESC,
		shortdesc = STRINGS.PROGRAMS.SHADE.SHORT_DESC,
		tipdesc = STRINGS.PROGRAMS.SHADE.TIP_DESC,

		icon = "gui/icons/programs_icons/icon-program-shade.png",
		icon_100 = "gui/icons/programs_icons/store_icons/StorePrograms_shade.png",
		cpu_cost = 1,
		cooldown = 0,
		maxCooldown = 2,
		value = 500,

		onSpawnAbility = function( self, sim )
			sim:addTrigger( simdefs.TRG_START_TURN, self )		
		end,

		onTrigger = function( self, sim, evType, evData )
			local player = sim:getCurrentPlayer()
			if self.used then 
				for _, unit in pairs(sim:getAllUnits()) do
                    if unit:getModifiers():remove( "shade" ) then
                        sim:refreshUnitLOS( unit )
						sim:dispatchEvent( simdefs.EV_UNIT_REFRESH, { unit = unit } )
					end
				end
				self.used = false 
			end

			if evType == simdefs.TRG_START_TURN and evData:isPC() then
				if self.cooldown > 0 then 
					self.cooldown = self.cooldown - 1
				end 
			end
		end,

		executeAbility = function( self, sim )
			self.used = true 
			local player = sim:getCurrentPlayer()
            self:useCPUs( sim )
			for _, unit in pairs(sim:getAllUnits()) do
				if unit:getPlayerOwner() ~= player and unit:getTraits().isGuard and unit:getTraits().hasSight then
                    unit:getModifiers():add( "LOSrange", "shade", modifiers.ADD, -3 )
                    sim:refreshUnitLOS( unit )
				end
			end
			self.cooldown = self.maxCooldown
		end,

	},

    
	leash = util.extend( DEFAULT_ABILITY ) 
	{
		name = STRINGS.PROGRAMS.LEASH.NAME,
		desc = STRINGS.PROGRAMS.LEASH.DESC,
		huddesc = STRINGS.PROGRAMS.LEASH.HUD_DESC,
		shortdesc = STRINGS.PROGRAMS.LEASH.SHORT_DESC,
		tipdesc = STRINGS.PROGRAMS.LEASH.TIP_DESC,
		passive = true,

		icon = "gui/icons/programs_icons/Programs0014.png",
		icon_100 = "gui/icons/programs_icons/store_icons/StorePrograms_0014.png",
		value = 500,

		canUseAbility = function( self, sim )
			return false 	
		end,

		onSpawnAbility = function( self, sim, player )
            player:getTraits().controlTicks = (player:getTraits().controlTicks or 0) + 1
		end,
	},

	dataBlast = util.extend( DEFAULT_ABILITY )
	{
		name = STRINGS.PROGRAMS.DATA_BLAST.NAME,
		desc = STRINGS.PROGRAMS.DATA_BLAST.DESC,
		huddesc = STRINGS.PROGRAMS.DATA_BLAST.HUD_DESC,
		shortdesc = STRINGS.PROGRAMS.DATA_BLAST.SHORT_DESC,
		tipdesc = STRINGS.PROGRAMS.DATA_BLAST.TIP_DESC,
 
		icon = "gui/icons/programs_icons/Programs0015.png",
		icon_100 = "gui/icons/programs_icons/store_icons/StorePrograms_0015.png",
		cpu_cost = 3,
        range = 5,
        value = 400,

		acquireTargets = function( self, targets, game, sim, unit )
            local targetHandler = targets.areaCellTarget( game, self.range, sim, sim:getPC() )
            targetHandler:setUnitPredicate(
                function( u )
                    return mainframe.canBreakIce( sim, u, self )
                end )
            targetHandler:setHiliteColor( { 0.33, 0.33, 0.33, 0.33 } )
            return targetHandler
		end, 

        startTargeting = function( self, sim, unit, userUnit, targetCell )
    		MOAIFmodDesigner.playSound( "SpySociety/Actions/mainframe_datablast_select" )
        end,

        getTargetUnits = function( self, sim, cellx, celly )
            local cells = simquery.rasterCircle( sim, cellx, celly, self.range )
            local units = {}
            for i, x, y in util.xypairs( cells ) do
                local cell = sim:getCell( x, y )
                if cell then
                    for _, cellUnit in ipairs(cell.units) do
                        if sim:getCurrentPlayer():getLastKnownCell( sim, x, y ) ~= nil then
                            if mainframe.canBreakIce( sim, cellUnit, self ) then
                                table.insert( units, cellUnit )
                            end
                        end
                    end
                end
            end
            return units
        end,

		executeAbility = function( self, sim, unit, userUnit, targetCell )
			local player = sim:getCurrentPlayer()
            local currentProgram = player:getEquippedProgram()

    		player:equipProgram( sim, self:getID() )

			local cellx, celly = unpack(targetCell)
            sim:dispatchEvent( simdefs.EV_PLAY_SOUND, "SpySociety/Actions/mainframe_datablast_use" )

            local targetUnits = self:getTargetUnits( sim, cellx, celly )

           	sim:dispatchEvent( simdefs.EV_OVERLOAD_VIZ, {x = cellx, y = celly, units = targetUnits, range = self.range } )		

            local daemonUnits = {}
            for _, unit in ipairs(targetUnits) do
                -- Keep track of daemons that *Would* be invoked, so that we can invoke them only after
                -- everything has been broken.
                local daemonProgram = unit:getTraits().mainframe_program

                unit:getTraits().mainframe_program = nil
                mainframe.breakIce( sim, unit, 1 )
                unit:getTraits().mainframe_program = daemonProgram

                if daemonProgram and unit:getTraits().mainframe_ice <= 0 then
                    table.insert( daemonUnits, unit )
                end
	        end

            self:useCPUs( sim )
    		player:equipProgram( sim, currentProgram and currentProgram:getID() )

            for i, unit in ipairs( daemonUnits ) do
                mainframe.invokeDaemon( sim, unit )
            end
		end,
	},

	esp = util.extend( DEFAULT_ABILITY )
	{

		name = STRINGS.PROGRAMS.ESP.NAME,
		desc = STRINGS.PROGRAMS.ESP.DESC,
		huddesc = STRINGS.PROGRAMS.ESP.HUD_DESC,
		shortdesc = STRINGS.PROGRAMS.ESP.SHORT_DESC,
		tipdesc = STRINGS.PROGRAMS.ESP.TIP_DESC,

		icon = "gui/icons/programs_icons/Program0018.png",
		icon_100 = "gui/icons/programs_icons/store_icons/StorePrograms_program18.png",
		cpu_cost = 3,
		credit_cost = 0,
		value = 300,

		canUseAbility = function( self, sim, abilityOwner )
			local player = sim:getCurrentPlayer()
			if player ~= abilityOwner or player == nil then
				return false
			end
			--Are there any untagged guards left? 
			local taggedguards = true
			for _, unit in pairs( sim:getAllUnits() ) do 	
				if sim:canPlayerSeeUnit( player, unit ) then
					if unit:getPlayerOwner() ~= player and unit:getTraits().isGuard and not unit:getTraits().tagged then 
						taggedguards = false 
					end
				end
			end

			if taggedguards then 
				return false, STRINGS.UI.REASON.NO_TAGGED_GUARDS
			end

			return DEFAULT_ABILITY.canUseAbility( self, sim, abilityOwner )
		end,

		executeAbility = function( self, sim, unit, userUnit, targetCell )
			local player = sim:getCurrentPlayer()
				
			local guards = {}

			sim:dispatchEvent( simdefs.EV_PLAY_SOUND, "SpySociety/Actions/mainframe_wisp_activate" )
			-- sound for spawning the tag effect on the guards "SpySociety/Actions/mainframe_wisp_reveal"

			for _, unit in pairs( sim:getAllUnits() ) do 
				if sim:canPlayerSeeUnit( player, unit ) then
					if unit:getPlayerOwner() ~= player and unit:getTraits().isGuard and not unit:getTraits().tagged then 										
				    	unit:setTagged()				    	
				    	sim:dispatchEvent( simdefs.EV_UNIT_TAGGED, {unit = unit} )
					end
				end
			end
		
            self:useCPUs( sim )
		end,

	},	

	pwr_manager = util.extend( DEFAULT_ABILITY )
	{
		name = STRINGS.PROGRAMS.PWR_MANAGER.NAME,
		desc = STRINGS.PROGRAMS.PWR_MANAGER.DESC,
		huddesc = STRINGS.PROGRAMS.PWR_MANAGER.HUD_DESC,
		shortdesc = STRINGS.PROGRAMS.PWR_MANAGER.SHORT_DESC,
		tipdesc = STRINGS.PROGRAMS.PWR_MANAGER.TIP_DESC,
 
 		passive = true,

		icon = "gui/icons/programs_icons/Program0019.png",
		icon_100 = "gui/icons/programs_icons/store_icons/StorePrograms_0019.png",
        value = 400,

		canUseAbility = function( self, sim )
			return false 	
		end,

		onSpawnAbility = function( self, sim, player )
            if not player:getTraits().extraStartingPWR then
            	player:getTraits().extraStartingPWR = 0
            end
            player:getTraits().extraStartingPWR = player:getTraits().extraStartingPWR + 4
		end,
	},

	pwr_manager_2 = util.extend( DEFAULT_ABILITY )
	{
		name = STRINGS.PROGRAMS.PWR_MANAGER_2.NAME,
		desc = STRINGS.PROGRAMS.PWR_MANAGER_2.DESC,
		huddesc = STRINGS.PROGRAMS.PWR_MANAGER_2.HUD_DESC,
		shortdesc = STRINGS.PROGRAMS.PWR_MANAGER_2.SHORT_DESC,
		tipdesc = STRINGS.PROGRAMS.PWR_MANAGER_2.TIP_DESC,
 		
 		passive = true,

		icon = "gui/icons/programs_icons/Program0019.png",
		icon_100 = "gui/icons/programs_icons/store_icons/StorePrograms_0019.png",
        value = 550,

		canUseAbility = function( self, sim )
			return false 	
		end,

		onSpawnAbility = function( self, sim, player )
            if not player:getTraits().extraStartingPWR then
            	player:getTraits().extraStartingPWR = 0
            end
            player:getTraits().extraStartingPWR = player:getTraits().extraStartingPWR + 6
		end,
	},	


	taurus = util.extend( DEFAULT_ABILITY )
	{
		name = STRINGS.PROGRAMS.TAURUS.NAME,
		desc = STRINGS.PROGRAMS.TAURUS.DESC,
		huddesc = STRINGS.PROGRAMS.TAURUS.HUD_DESC,
		shortdesc = STRINGS.PROGRAMS.TAURUS.SHORT_DESC,
		tipdesc = STRINGS.PROGRAMS.TAURUS.TIP_DESC,

		icon = "gui/icons/programs_icons/Program0020.png",
		icon_100 = "gui/icons/programs_icons/store_icons/StorePrograms_0020.png",
		cpu_cost = 2,
		break_firewalls = 0, 
		equip_program = true, 
		equipped = false,
		daemon_killer = true,   -- because it targets a daemon		
		value = 400,

		canUseAbility = function( self, sim, abilityOwner )
			
			--Are there any places the Daemon can go? 
			local possibleUnits = {}
			for _, unit in pairs( sim:getAllUnits() ) do
				if unit:getTraits().mainframe_iceMax and unit:getTraits().mainframe_iceMax > 0 and unit:getTraits().mainframe_ice and unit:getTraits().mainframe_ice > 0 and unit:getPlayerOwner() ~= sim:getPC() and not unit:getTraits().mainframe_program then
					table.insert( possibleUnits, unit )		
				end
			end

			if #possibleUnits == 0 then 
				return false, STRINGS.UI.REASON.NO_ROOM_FOR_DAEMON
			end

			return DEFAULT_ABILITY.canUseAbility( self, sim, abilityOwner )
		end,

		executeAbility = function( self, sim, targetUnit )
            self:useCPUs( sim )

            if targetUnit:getTraits().daemonHost then
                -- Its a daemon host... punt it somewhere else.
                local host = sim:getUnit( targetUnit:getTraits().daemonHost )
                sim:moveDaemon( host )

            else
			    local possibleUnits = {}
			    for _, unit in pairs( sim:getAllUnits() ) do
				    if unit:getTraits().mainframe_item and unit:getTraits().mainframe_ice > 0 and unit:getPlayerOwner() ~= sim:getPC() and not unit:getTraits().mainframe_program then
					    table.insert( possibleUnits, unit )		
				    end
			    end

			    for k=1,1,1 do 
				    if #possibleUnits > 0 then 
					    local index = sim:nextRand(1, #possibleUnits)
					    local unit = possibleUnits[ index ]
					    table.remove( possibleUnits, index )

					    local program = targetUnit:getTraits().mainframe_program 
					    targetUnit:getTraits().mainframe_program = nil

					    sim:dispatchEvent( simdefs.EV_PLAY_SOUND, "SpySociety/Actions/mainframe_daemonmove")
					    sim:dispatchEvent( simdefs.EV_UNIT_UPDATE_ICE, { unit = unit, ice = unit:getTraits().mainframe_ice, delta = 0, refreshAll = true} )

					    sim:dispatchEvent(simdefs.EV_MAINFRAME_MOVE_DAEMON, {source=targetUnit, target=unit})
					
					    unit:getTraits().mainframe_program = program
					    sim:dispatchEvent( simdefs.EV_PLAY_SOUND, "SpySociety/Actions/mainframe_daemon_spawned" )
					    sim:dispatchEvent( simdefs.EV_UNIT_UPDATE_ICE, { unit = unit, ice = unit:getTraits().mainframe_ice, delta = 0} )									
				    end 
			    end
            end
			
		end,
	},

}

return mainframe_abilities
